-- TBD: Have configuration follow the action name?
ABT = LibStub("AceAddon-3.0"):NewAddon("ButtonTimers", "AceConsole-3.0", "AceEvent-3.0")
local L = LibStub("AceLocale-3.0"):GetLocale("ButtonTimers", "enUS", true)
ABT_MaxButton = 12
ABT_MaxBar = 4
ABT_Widgets = "ABTWid"
local ABT_Bars = {}
local ABT_Buttons = {}
local currentBar = 1
local currentButton = 0
local outer_margin = 8
local ABT_TARGET = 1
local ABT_FOCUS = 2
local ABT_PLAYER = 3
local ABT_MOUSEOVER = 4
local ABT_PARTY = 5
local ABT_TARGET_NAMES = { [ABT_TARGET]="target", [ABT_FOCUS]="focus", [ABT_PLAYER]="player", [ABT_PARTY]="party" }
local ABT_AURA = 1
local ABT_COOLDOWN = 2
local ABT_NONE = 3
local ABT_VERTICAL = 1
local ABT_HORIZONTAL = 2
local ABT_FULL = 1
local ABT_FIXED = 2
local TimeSinceLastUpdate = nil
local ABT_UpdateInterval = .05


function NS (arg1)
	if arg1 == nil then
		return "nil"
	elseif type(arg1) == "table" then
		local myval = "{"
		for i, j in pairs(arg1) do
			myval = myval.." ["..NS(i).."]="..NS(j)..","
		end
		myval = myval.."}"
		return myval
	else
		return tostring(arg1)
	end
end

function ABT:DebugTrace (arg1, enter)
	if ABT.db and ABT.db.char.showTrace == true then
		if (enter) then
			DEFAULT_CHAT_FRAME:AddMessage("ABT:"..arg1.." enter")
		else
			DEFAULT_CHAT_FRAME:AddMessage("ABT:"..arg1.." exit")
		end
	end
end

function ABT:DebugPrint (arg1)
	if ABT.db.char.showDebug == true then
		DEFAULT_CHAT_FRAME:AddMessage("ABT:"..arg1)
	end
end

function deepcopy(object)
    local lookup_table = {}
    local function _copy(object)
        if type(object) ~= "table" then
            return object
        elseif lookup_table[object] then
            return lookup_table[object]
        end  -- if
        local new_table = {}
        lookup_table[object] = new_table
        for index, value in pairs(object) do
            new_table[_copy(index)] = _copy(value)
        end  -- for
        return setmetatable(new_table, getmetatable(object))
    end  -- function _copy
    return _copy(object)
end  -- function deepcopy

-- parse a comma separated list into a table
function parseList (list)
	local _table = {}
	local idx = 1
	if list then
		for value in string.gmatch(list, "[^\n\r,]+") do
			tinsert(_table, value)
		end
	end
	return _table
end
function getKey(info)
	local key
	if #info == 3 then
		key = info[#info-2]..info[#info-1]..info[#info]
	elseif #info == 2 then
		key = info[#info-1]..info[#info]
	else
		key = info[#info]
	end
	return key
end

function ABT:GetValue (barIdx, buttonIdx, field)
	local key = nil
	if buttonIdx and barIdx and field then
		key = "bar"..barIdx.."button"..buttonIdx..field
	elseif barIdx and field then
		key = "bar"..barIdx..field
	elseif field then
		key = field
	end
	if key then
		return ABT.db.char[key]
	end
end

function ABT:SetValue (barIdx, buttonIdx, field, value)
	local key = nil
	if buttonIdx and barIdx and field then
		key = "bar"..barIdx.."button"..buttonIdx..field
	elseif barIdx and field then
		key = "bar"..barIdx..field
	elseif field then
		key = field
	end
	if key then
		ABT.db.char[key] = value
	end
end

function ABT:GetButtonLabel(info)
	local barIdx, buttonIdx = getIndex (info)
	return ABT:GetButtonName (barIdx, buttonIdx)
end

function ABT:GetButtonHidden(info)
	local barIdx, buttonIdx = getIndex (info)
	if barIdx and buttonIdx then
		return buttonIdx > ABT:GetValue (barIdx, nil, "buttonCount")
	else
		return false
	end
end

function getIndex(info)
	local buttonIdx = nil
	local barIdx = nil
	local i
	local j
	for i = 0,#info do
		if info[#info-i] then
			local a = string.match(info[#info-i], "button(.+)")
			if a then buttonIdx = tonumber(a) break end
		end
	end
	for i = 0,#info do
		if info[#info-i] then
			local a = string.match(info[#info-i], "bar(.+)")
			if a then barIdx = tonumber(a) break end
		end
	end
	return barIdx, buttonIdx
end

function newGetIndex(info)
	local buttonIdx = nil
	local barIdx = nil
	local i = 0
	
	local field = info[#info]
	local a = string.match(info[#info-i], "button(.+)")
	if a then
		buttonIdx = tonumber(a)
		i = i + 1
	end
	a = string.match(info[#info-i], "bar(.+)")
	if a then
		barIdx = tonumber(a)
	end
	return barIdx, buttonIdx, field
end

function ABT:getFunc(info)
	local key = getKey(info)
--	ABT:DebugPrint ("get "..key)
	return (self.db.char[key])
end

function ABT:setFunc(info, value)
	local key = getKey(info)
--	ABT:DebugPrint ("set "..key)
	self.db.char[key] = value
	local barIdx, buttonIdx = getIndex(info)
	if barIdx ~= nil then   -- button option or bar option
		ABT:SetBar (barIdx)
	else                                             -- general option
		for barIdx = 1, ABT_MaxBar do
			ABT:SetBar (barIdx)
		end
	end
	ABT:GetDebuffs()
end

local buttonOptions = {
	type = "group",
	name = function(info) return ABT:GetButtonLabel(info) end,
	hidden = function(info) return ABT:GetButtonHidden(info) end,
	desc = "Options for each button",
	childGroups = "tree",
	inline = false,
	args = {
		target = {
			order = 2,
			type = "select",
			name = L["Target"],
			desc = L["The target on which this button will operate."],
			values = { [ABT_TARGET]=L["Target"], [ABT_FOCUS]=L["Focus"], [ABT_PLAYER]=L["Player"], [ABT_PARTY]=L["Party/Raid Member"] },
			style = "radio",
			set = "setFunc",
			get = "getFunc",
		},
		playerTarget = {
			order = 3,
			type = "select",
			hidden = function (info)  return ABT.db.char[info[#info-2]..info[#info-1].."target"] ~= ABT_PARTY  end,
			name = L["Party/Raid Member"],
			desc = L["The party / raid member to target"],
			values = function () return ABT:GetPartyList() end,
			style = "radio",
			set = "setFunc",
			get = "getFunc",
		},
		timerType = {
			type = "select",
			order = 10,
			name = L["Type"],
			values = { [ABT_AURA]=L["Aura"], [ABT_COOLDOWN]=L["Cooldown"], [ABT_NONE]=L["None"]},
			style = "radio",
			set = "setFunc",
			get = "getFunc",
		},
		barType = {
			type = "select",
			order = 20,
			name = L["Timer Type"],
			desc = L["Choose full time to see the full length of the timer, choose fixed time to see a set maximum time."],
			values = { [ABT_FULL]=L["Full time"], [ABT_FIXED]=L["Fixed time"] },
			style = "radio",
			set = "setFunc",
			get = "getFunc",
		},
		barTime = {
			order = 30,
			type = "range",
			hidden = function (info)  return ABT.db.char[info[#info-2]..info[#info-1].."barType"] ~= ABT_FIXED  end,
			name = L["Bar Time"],
			desc = L["Max time displayed on bar"],
			min = 1,
			max = 100,
			step = 1,
			bigStep = 1,
			set = "setFunc",
			get = "getFunc",
		},
		showOthers = {
			order = 40,
			type = "toggle",
			hidden = function (info)  return ABT.db.char[info[#info-2]..info[#info-1].."timerType"] ~= ABT_AURA  end,
			name = L["Show others spells"],
			desc = L["Show this spell on the target even if someone else cast it."],
			set = "setFunc",
			get = "getFunc",
		},
		auras = {
			order = 50,
			hidden = function (info)  return ABT.db.char[info[#info-2]..info[#info-1].."timerType"] ~= ABT_AURA  end,
			type = "input",
			name = L["Other Auras"],
			desc = L["Add any additional auras you want the timer to show as a comma separated list."],
			multiline = true,
			set = "setFunc",
			get = "getFunc",
		},
		spell = {
			order = 60,
			hidden = function (info)  return ABT.db.char[info[#info-2]..info[#info-1].."timerType"] ~= ABT_COOLDOWN  end,
			type = "input",
			name = L["Cooldown Spell"],
			desc = L["If you would like to see the cooldown for a spell other than the one on the button, enter it here."],
			set = "setFunc",
			get = "getFunc",
		},
		timerColor = {
			order = 70,
            name = L["Bar Color"],
            desc = L["Bar color for timers"],
            type = "color",
			hasAlpha = true,
            set = function (info, r, g, b, a) return ABT:setFunc (info, {r, g, b, a}) end,
            get = function (info) return unpack(ABT:getFunc(info)) end
        }
	}
}

function ABT:GetButtonOptions (button)
	local _options = deepcopy (buttonOptions)
	_options.order = button
	return _options
end

local barOptions = {
	type = "group",
	name = "Bar",
--	handler = ABT,
	desc = "Options for the bar",
	childGroups = "tree",
	inline = false,
	args = {
		enabled = {
			order = 10,
			type = "toggle",
			name = L["Bar Enabled"],
			desc = L["Enable this bar"],
			set = "setFunc",
			get = "getFunc",
		},
		buttonCount = {
			order = 20,
			type = "range",
			name = L["Button Count"],
			desc = L["Number of Buttons on bar"],
			min = 1,
			max = 12,
			step = 1,
			bigStep = 1,
			set = "setFunc",
			get = "getFunc",
		},
		actionOffset = {
			order = 30,
			type = "range",
			name = L["Action Offset"],
			desc = L["Move this slider to change which action slots are shown on the bar."],
			min = 0,
			max = 119,
			step = 1,
			bigStep = 1,
			set = "setFunc",
			get = "getFunc",
		},	
		orientation = {
			order = 40,
			type = "select",
			name = L["Orientation"],
			values = { [ABT_VERTICAL]=L["Vertical"], [ABT_HORIZONTAL]=L["Horizontal"] },
			style = "radio",
			set = "setFunc",
			get = "getFunc",
		},
		locked = {
			order = 50,
			type = "toggle",
			name = L["Bar Locked"],
			desc = L["Lock bar in place"],
			set = "setFunc",
			get = "getFunc",
		},
		spacing = {
			order = 60,
			type = "range",
			name = L["Button Spacing"],
			desc = L["The space between the buttons"],
			min = 0,
			min = 0,
			max = 16,
			step = 1,
			bigStep = 1,
			set = "setFunc",
			get = "getFunc",
		},
		scale = {
			order = 70,
			type = "range",
			name = L["Bar Scale"],
			desc = L["Allows you to scale the bar size"],
			min = .1,
			max = 2,
			step = .01,
			bigStep = .01,
			set = "setFunc",
			get = "getFunc",
		},
		length = {
			order = 80,
			type = "range",
			name = L["Bar Length"],
			desc = L["Length of bar in pixels"],
			min = 1,
			max = 600,
			step = 1,
			bigStep = 1,
			set = "setFunc",
			get = "getFunc",
		},
		button1 = ABT:GetButtonOptions(1),
		button2 = ABT:GetButtonOptions(2),
		button3 = ABT:GetButtonOptions(3),
		button4 = ABT:GetButtonOptions(4),
		button5 = ABT:GetButtonOptions(5),
		button6 = ABT:GetButtonOptions(6),
		button7 = ABT:GetButtonOptions(7),
		button8 = ABT:GetButtonOptions(8),
		button9 = ABT:GetButtonOptions(9),
		button10 = ABT:GetButtonOptions(10),
		button11 = ABT:GetButtonOptions(11),
		button12 = ABT:GetButtonOptions(12),
	}
}

function ABT:GetBarOptions (bar)
	local _options = deepcopy (barOptions)
	_options.name = "Bar ".. bar
	_options.order = bar
	return _options
end

local options = {
    name = "ButtonTimers",
    handler = ABT,
    type = 'group',
	childGroups = "tree",
    args = {
		enable={
			order = 1,
			name= L["Enable"],
			desc= L["Enables / disables the addon"],
			type= "toggle",
			set = "setFunc",
			get = "getFunc",
		},
	    showDebug = {
			order = 2,
			hidden = true,
            name = L["ShowDebug"],
            desc = L["Show debug messages"],
            type = "toggle",
            set = "setFunc",
            get = "getFunc",
        },
	    showTrace = {
			order = 2,
			hidden = true,
            name = L["ShowTrace"],
            desc = L["Show trace messages"],
            type = "toggle",
            set = "setFunc",
            get = "getFunc",
        },
		showBackground = {
			order = 1,
			hidden = true,
            name = L["ShowBackground"],
            desc = L["Show background bar"],
            type = "toggle",
            set = "setFunc",
            get = "getFunc",
        },
		bar1 = ABT:GetBarOptions(1),
		bar2 = ABT:GetBarOptions(2),
		bar3 = ABT:GetBarOptions(3),
		bar4 = ABT:GetBarOptions(4),
    },
}

function ABT:InitDefaults()
	local defaults = {
		char = {}
	}
	defaults.char.enable = true
	defaults.char.showDebug = false
	defaults.char.showTrace = false
	defaults.char.showBackground = false
	for bar = 1, ABT_MaxBar do
		defaults.char["bar"..bar.."enabled"] = true
		defaults.char["bar"..bar.."locked"] = false
		defaults.char["bar"..bar.."orientation"] = 1
		defaults.char["bar"..bar.."scale"] = 1
		defaults.char["bar"..bar.."length"] = 300
		defaults.char["bar"..bar.."spacing"] = 8
		defaults.char["bar"..bar.."buttonCount"] = 12
		defaults.char["bar"..bar.."actionOffset"] = 0
		defaults.char["bar"..bar.."auraColor"] = { 0, 1, 1, .5 }
		defaults.char["bar"..bar.."cooldownColor"] = { 0, 1, 0, .5 }
		for button = 1, ABT_MaxButton do
			defaults.char["bar"..bar.."button"..button.."target"] = ABT_TARGET
			defaults.char["bar"..bar.."button"..button.."timerType"] = ABT_AURA
			defaults.char["bar"..bar.."button"..button.."barType"] = ABT_FULL
			defaults.char["bar"..bar.."button"..button.."barTime"] = 20
			defaults.char["bar"..bar.."button"..button.."showOthers"] = false
			defaults.char["bar"..bar.."button"..button.."auras"] = nil
			defaults.char["bar"..bar.."button"..button.."spell"] = nil
			defaults.char["bar"..bar.."button"..button.."timerColor"] = { 0, 1, 1, .5 }
		end
	end
	return defaults
end

function ABT:OnInitialize()
  -- Code that you want to run when the addon is first loaded goes here.
	ABT:DebugTrace ("OnInitialize", true)
	ABT:RegisterChatCommand("buttontimers", "SlashProcessorFunc")
	-- overwrites bartender, doh
	--	ABT:RegisterChatCommand("bt", "SlashProcessorFunc")
    LibStub("AceConfig-3.0"):RegisterOptionsTable("ButtonTimers", options)
    self.optionsFrame = LibStub("AceConfigDialog-3.0"):AddToBlizOptions("ButtonTimers", "ButtonTimers")
    self.db = LibStub("AceDB-3.0"):New("ButtonTimersDB", ABT:InitDefaults(), "Default")
	ABT:MakeWidgets()
	ABT:DebugTrace ("OnInitialize")

	-- Binding Variables
	for bar=1,ABT_MaxBar do
		setglobal("BINDING_HEADER_ButtonTimersBar"..bar, "ButtonTimers Bar "..bar);
		for button=1, ABT_MaxButton do
			local buttonName = ABT_Widgets.."bar"..bar.."button"..button
			setglobal("BINDING_NAME_CLICK "..buttonName..":LeftButton", "Button "..button);
		end
	end
	ABT:ButtonUpdate()
end

function ABT:OnEnable()
	ABT:DebugTrace ("OnEnable", true)
    -- Called when the addon is enabled
	self:RegisterEvent ("UNIT_AURA", "GetDebuffs")
	self:RegisterEvent ("PLAYER_TARGET_CHANGED", "GetDebuffs")
	self:RegisterEvent ("PLAYER_FOCUS_CHANGED", "GetDebuffs")
	self:RegisterEvent ("ACTIONBAR_UPDATE_COOLDOWN", "GetCooldowns")
	self:RegisterEvent ("ACTIONBAR_UPDATE_USABLE", "GetCooldowns")
	self:RegisterEvent ("PARTY_MEMBERS_CHANGED", "UpdatePlayerTargets")
	self:RegisterEvent ("ACTIONBAR_SLOT_CHANGED", function (event, arg1) 
													ActionButton_OnEvent(event) ABT:ButtonUpdate() 
													end)
	ABT:DebugTrace ("OnEnable")
end

function ABT:ButtonUpdate ()
	ABT:GetCooldowns()
	ABT:GetDebuffs()
end

function ABT:OnDisable()
	ABT:DebugTrace ("OnDisable", true)
    -- Called when the addon is disabled
	ABT:DebugTrace ("OnDisable")
end

function ABT:SlashProcessorFunc(input)
	ABT:DebugTrace ("SlashProcessorFunc", true)
  -- Process the slash command ('input' contains whatever follows the slash command)
	ABT:DebugPrint ("Received command "..input)
	if (input == "menu" or input == "config") then
		InterfaceOptionsFrame_OpenToCategory("ButtonTimers")
	elseif (input == "reset") then
		for barIdx = 0, ABT_MaxButton do
			local bar = ABT_Bars[barIdx]
			if bar then
				bar:ClearAllPoints()
				bar:SetPoint("CENTER")
			end
		end
	elseif (input == "toggleDebug") then
		if ABT.db.char.showDebug then
			ABT:DebugPrint ("turning debug messages off")
			ABT.db.char.showDebug = false
		else
			ABT.db.char.showDebug = true
			ABT:DebugPrint ("turning debug messages on")
		end
	elseif (input == "toggleTrace") then
		if ABT.db.char.showTrace then
			ABT:DebugPrint ("turning trace messages off")
			ABT.db.char.showTrace = false
		else
			ABT.db.char.showTrace = true
			ABT:DebugPrint ("turning trace messages on")
		end
	else
		DEFAULT_CHAT_FRAME:AddMessage("/buttonTimers <usage>")
		DEFAULT_CHAT_FRAME:AddMessage("/buttonTimers menu or config - Show config options")
		DEFAULT_CHAT_FRAME:AddMessage("/buttonTimers reset - Reset windows to default positions")		
	end
	ABT:DebugTrace ("SlashProcessorFunc")
end

-- update the bars
function ABT:OnUpdate (self, elapsed)
	if self:IsShown() and ABT:GetValue(nil, nil, "enable") then
		if self.TimeSinceLastUpdate == nil then 
			self.TimeSinceLastUpdate = elapsed
		end
		self.TimeSinceLastUpdate = self.TimeSinceLastUpdate + elapsed; 	
		if (self.TimeSinceLastUpdate > ABT_UpdateInterval) then
			local currentTime = GetTime()
			self.TimeSinceLastUpdate = 0;

			if self.expirationTime and self.expirationTime > currentTime then
				local remaining = self.expirationTime - currentTime;
				if remaining > 10 then
					self.countdown:SetFormattedText(SecondsToTimeAbbrev(remaining))
				else
					self.countdown:SetFormattedText("%.1f", remaining)
				end
				self.status:SetValue(remaining)
			else
				self.expirationTime = nil
				self.countdown:SetText(nil)
				self.spellName:SetText(nil)
				self.status:SetValue(0)
				self.status:Hide()
			end
		end
	end
end

function ABT:GetButtonName (barIdx, buttonIdx)
	local buttonName
	local button = ABT_Buttons[barIdx][buttonIdx]
	local actionSlot = button:GetAttribute("action")	
	local actiontype, id, subType
	if actionSlot then
		actiontype, id, subType = GetActionInfo(actionSlot)
--		ABT:DebugPrint ("type="..NS(actiontype).." id="..NS(id).."subType="..NS(subType))
	end
	if actiontype == "spell" and id > 0 then
		local spellName = GetSpellName (id, BOOKTYPE_SPELL)
--		ABT:DebugPrint ("spellName = "..NS(spellName))
		if spellName then
			buttonName = spellName
		end
	elseif actiontype == "macro" then
		buttonName = GetActionText(actionSlot)
	elseif actiontype == "item" then
		buttonName = GetItemInfo(id)
	end
	if buttonName == nil then
		buttonName = "bar"..barIdx.."button"..buttonIdx
	end
	return buttonName
end

--
-- Help, does this really need to do all buttons for each call?
-- couldn't we just update cooldowns on buttons that gave the action?
function ABT:GetCooldowns (self, arg1)
	ABT:DebugTrace ("GetCooldowns", true)
	ABT:InitButtonLoop()
	repeat 
		local barIdx, buttonIdx = ABT:GetNextButton()
		if barIdx and buttonIdx then
			local key = "bar"..barIdx.."button"..buttonIdx
			local timerType = ABT:GetValue (barIdx, buttonIdx, "timerType")
			if timerType == ABT_COOLDOWN then -- cooldown timer
				local otherSpell = ABT:GetValue (barIdx, buttonIdx, "spell")
				local button = ABT_Buttons[barIdx][buttonIdx]
				local slot = button:GetAttribute ("action")

				local start, duration, enabled
				if otherSpell then
					start, duration, enabled = GetSpellCooldown (otherSpell)
				else
					start, duration, enabled = GetActionCooldown (slot)
				end
				if start and duration then
					-- we have a cooldown to track
					if start > 0 and duration > 1.5 then
						button.expirationTime = start + duration
						if otherSpell then
							button.spellName:SetText(otherSpell)
						else
							-- find the name of the spell, so we can put it on the timer
							spellName = ABT:GetButtonName (barIdx, buttonIdx)
							if spellName then
								button.spellName:SetText(spellName)
							end
						end
						local barTime
						if ABT:GetValue (barIdx, buttonIdx, "barType") == ABT_FIXED then
							barTime = ABT:GetValue (barIdx, buttonIdx, "barTime")
						else
							barTime = duration
						end
						button.status:SetMinMaxValues (0, barTime)
						button.status:Show()
					-- we have a cooldown timer running, but the cooldown is no longer
					-- active. Reset it. If it's less than the GCD, though, just let it run out.
					elseif button.expirationTime and (button.expirationTime - GetTime()) > 1.5 then
						button.expirationTime = nil
					end
				end
			end
		end
	until buttonIdx == nil
	ABT:DebugTrace ("GetCooldowns")
end

function ABT:GetDebuffs(event, arg1)
	ABT:DebugTrace ("GetDebuffs", true)
	local unit = nil
--	ABT:DebugPrint (NS(event).." "..NS(arg1))
	if (event == "UNIT_AURA" and arg1 == "target") or event == "PLAYER_TARGET_CHANGED" then
		unit = "target"
	elseif (event == "UNIT_AURA" and arg1 == "focus") or event == "PLAYER_FOCUS_CHANGED" then
		unit = "focus"
	elseif (event == "UNIT_AURA") then
		unit = arg1
	end
--	if unit then
		-- loop over buttons looking for aura timer buttons matching the target
		local barIdx = nil
		local buttonIdx = nil
		ABT:InitButtonLoop()
		repeat 
			barIdx, buttonIdx = ABT:GetNextButton()
--			ABT:DebugPrint ("bar="..NS(barIdx).." button="..NS(buttonIdx))
			if barIdx and buttonIdx then
				local button = ABT_Buttons[barIdx][buttonIdx]
				local key = "bar"..barIdx.."button"..buttonIdx
				local timerType = ABT:GetValue (barIdx, buttonIdx, "timerType")
				local target = button:GetAttribute ("unit")
				if (target == nil) then target = "target" end
				local showOthers = ABT:GetValue (barIdx, buttonIdx, "showOthers")
				local buttonFound = false -- haven't yet found a match for this button
--				ABT:DebugPrint ("timerType="..NS(timerType).." target="..NS(target).." unit="..NS(unit))
				if timerType == ABT_AURA and (target == unit or unit == nil) then -- aura timer
					local effectList = parseList (ABT:GetValue (barIdx, buttonIdx, "auras"))
					ABT:DebugPrint ("button "..buttonIdx.." on bar "..barIdx.." looking for target "..NS(target).." buttonSpell="..NS(button.spell))
					-- build effects list
					if button.action then
						local actionType, actionId, actionSubType = GetActionInfo (button.action)
						if actionType == "spell" and actionId > 0 then
							local spellName, _ = GetSpellName ( actionId, BOOKTYPE_SPELL )
							effectList[100] = spellName
						end
					end
					-- now look for the aura for this button
					local effect
					for _, effect in pairs(effectList) do			
						ABT:DebugPrint ("Looking for "..effect.." on button "..button:GetName().." target="..NS(target))
						local name, rank, iconTexture, count, debuffType, duration, timeLeft, expirationTime, isMine
						timeLeft = nil
						isMine = nil
						expirationTime = nil
						name, rank, iconTexture, count, debuffType, duration, expirationTime, isMine = UnitDebuff (target, effect); -- 3.0

						if name == nil then -- wasn't found as a debuff, try it as a buff
							name, rank, iconTexture, count, debuffType, duration, expirationTime, isMine = UnitBuff (target, effect); -- 3.0
						end
						-- the effect was found on the mob, now lets see if it's ours
						-- use the effect if it is ours, or if the anyCaster flag is set and the button 
						-- was not previously found. This will prioritize our own spells for display if multiple
						-- matching effects are found
						if name ~= nil and (isMine or (showOthers and not buttonFound)) then
							ABT:DebugPrint ("found effect "..effect)
							local currenttime = GetTime()
							timeLeft = expirationTime-currenttime
							local timeStart = expirationTime - duration
							button.expirationTime = expirationTime -- time left on this effect
							ABT:DebugPrint ("expirationTime ="..button.expirationTime)
							local barTime
							if ABT:GetValue (barIdx, buttonIdx, "barType") == ABT_FIXED then
								barTime = ABT:GetValue (barIdx, buttonIdx, "barTime")
							else
								barTime = duration
							end
							button.status:SetMinMaxValues (0, barTime)
							button.status:Show()
							if target ~= "target" then
								button.spellName:SetText(effect..": "..NS(UnitName(target)))
							else
								button.spellName:SetText(effect)
							end
							buttonFound = true
						end
					end
					if not buttonFound then
						button.expirationTime = nil
					end
				end
			end
		until buttonIdx == nil
--	end
	ABT:DebugTrace ("GetDebuffs")
end

function ABT:InitButtonLoop()
	currentBar = 1
	currentButton = 0
end

function ABT:GetNextButton()
	while true do
		currentButton = currentButton + 1
		if currentButton > ABT_MaxButton then
			currentBar = currentBar + 1
			currentButton = 1
			if currentBar > ABT_MaxBar then
				return nil, nil
			end
		end
		local barEnabled = self.db.char["bar"..currentBar.."enabled"]
		local barButtonCount = self.db.char["bar"..currentBar.."buttonCount"]
		if barEnabled and currentButton <= barButtonCount then
			return currentBar, currentButton
		end
	end
end

function ABT:StartMoving (self)
	ABT:DebugTrace ("StartMoving", true)
	if (not InCombatLockdown() and not self.locked) then
		self:StartMoving()
	end
	ABT:DebugTrace ("StartMoving")
end

function ABT:StopMoving (self)
	ABT:DebugTrace ("StopMoving", true)
	self:StopMovingOrSizing ()
	local barIdx = self.barIdx
	local attach = { self:GetPoint() }
	ABT:SetValue (barIdx, nil, "attach", attach)
	ABT:DebugPrint ("stop moving bar="..barIdx.." "..NS(attach))
	ABT:DebugTrace ("StopMoving", true)
end

function ABT:MakeWidgets()
	ABT:DebugTrace ("MakeWidgets", true)
	for barIdx = 1, ABT_MaxBar do
		local barName = ABT_Widgets.."bar"..barIdx
		local bar = CreateFrame ("Frame", barName)
		ABT_Bars[barIdx] = bar
		bar.barIdx = barIdx
		-- this texture is handy for seeing where the frame is to drag it.
		-- only show it if the bar is draggable.
		bar.texture = bar:CreateTexture ()
		bar.texture:SetAllPoints(bar)
		bar.texture:SetTexture(1, 1, 1, .1)
		bar.texture:Hide()
		--
		-- setup for dragging
		bar:SetMovable (true)
		bar:SetClampedToScreen (true)
		bar:EnableMouse (true)
		bar:SetScript ("OnMouseDown", function (self) ABT:StartMoving (self) end)
		bar:SetScript ("OnMouseUp", function (self) ABT:StopMoving (self) end)
		bar:SetScript ("OnDragStop", function (self) ABT:StopMoving () end)

		ABT_Buttons[barIdx] = {}
		for buttonIdx = 1, ABT_MaxButton do
			local buttonName = barName.."button"..buttonIdx
			local button = CreateFrame ("CheckButton", buttonName , bar, "ABTButtonTemplate")
			ABT_Buttons[barIdx][buttonIdx] = button
			button:SetScript ("OnUpdate", function (self, elapsed) ActionButton_OnUpdate(self, elapsed) ABT:OnUpdate (self, elapsed) end)
			button.barIdx = barIdx
			button.buttonIdx = buttonIdx
			button.status = getglobal (buttonName.."StatusBar")
			button.spellName = getglobal (buttonName.."StatusBarSpellName")
			button.countdown = getglobal (buttonName.."StatusBarCountdown")
			button.status:Show()
		end
		ABT:SetBar(barIdx)
	end
	ABT:DebugTrace ("MakeWidgets")
end

function ABT:ButtonVisible (barIdx, buttonIdx)
	local buttonCount = ABT:GetValue (barIdx, nil, "buttonCount")
	return buttonIdx <= buttonCount
end

local scaleInit = {}

function ABT:SetBar (barIdx)
	ABT:DebugTrace ("SetBar"..barIdx, true)
	local width = 0
	local height = 0

	local buttonCount = ABT:GetValue(barIdx, nil, "buttonCount")
	local orientation = ABT:GetValue(barIdx, nil, "orientation")
	local scale = ABT:GetValue(barIdx, nil, "scale")
	local length = ABT:GetValue(barIdx, nil, "length")
	local margin = ABT:GetValue(barIdx, nil, "spacing")
	local actionOffset = ABT:GetValue(barIdx, nil, "actionOffset")
	local enabled = ABT:GetValue(barIdx, nil, "enabled") and ABT:GetValue(nil, nil, "enable")
	local locked = ABT:GetValue(barIdx, nil, "locked")
	local showBackground = ABT:GetValue(nil, nil, "showBackground")
	
	local bar = ABT_Bars[barIdx]

	-- bar enable
	if enabled then
		bar:Show()
	else
		bar:Hide()
		return
	end
	
	-- bar lockdown
	bar.locked = locked
	
	local attach = ABT:GetValue(barIdx, nil, "attach")
	ABT:DebugPrint ("read localtion from DB "..NS(attach))
	-- set bar position
	if (attach == nil) then
		attach = {"CENTER", nil, "CENTER", (44 * (barIdx-1)), 0}
		ABT:SetValue (barIdx, nil, "attach", attach)
		ABT:DebugPrint ("initialized attach="..NS(attach))
	end
 	bar:ClearAllPoints()
	bar:SetPoint(unpack(attach))

	-- scale option
	local oldscale = bar:GetScale()
	local newscale = scale  * .6

	if oldscale ~= newscale and attach then
		bar:SetScale(newscale)
		if scaleInit[barIdx] ~= nil then
			local point, parent, relpoint, xOfs, yOfs = unpack(attach)
			xOfs = xOfs * oldscale / newscale
			yOfs = yOfs * oldscale / newscale
			attach = {point, parent, relpoint, xOfs, yOfs}
			ABT:SetValue (barIdx, nil, "attach", attach)
			ABT:DebugPrint ("scaled DB attach="..NS(attach))
			bar:ClearAllPoints()
			bar:SetPoint(unpack(attach))
		else
			scaleInit[barIdx] = true		
		end
	end
	
	-- button count, target
	for buttonIdx = 1, ABT_MaxButton do
		local button = ABT_Buttons[barIdx][buttonIdx]
		if ABT:ButtonVisible (barIdx, buttonIdx) then
			button:Show()
			local targetIdx = ABT:GetValue(barIdx, buttonIdx, "target")
			local targetUnit = nil
			if targetIdx == ABT_PARTY then
				local playerTarget = ABT:GetValue(barIdx, buttonIdx, "playerTarget")
				targetUnit = ABT:FindPlayerUnitId (playerTarget)
			elseif targetIdx ~= ABT_TARGET then
				targetUnit = ABT_TARGET_NAMES[targetIdx]
			else
				targetUnit = nil
			end
			button:SetAttribute("unit", targetUnit)

			local statusbarTexture = "Interface\PaperDollInfoFrame\UI-Character-Skills-Bar"
			local statusbarTexture = "Interface\TargetingFrame\UI-StatusBar"
			local statusbarColor = ABT:GetValue (barIdx, buttonIdx, "timerColor")
			button.status:SetStatusBarTexture (statusbarTexture)
			button.status:SetStatusBarColor (unpack(statusbarColor))

			ABT:DebugPrint ("setting unit for "..button:GetName().." to "..NS(targetUnit).." targetIdx="..targetIdx)
			button:SetAttribute ("action", buttonIdx + actionOffset)
			ABT:DebugPrint ("Setting action on bar="..barIdx.." button="..buttonIdx.." to "..buttonIdx+actionOffset)
			height = button:GetHeight()
			width = button:GetWidth()
			local x, y, statusX, statusY
			if orientation == ABT_VERTICAL then
				x = outer_margin
				y = -1* (buttonIdx-1) * (height+margin) - outer_margin
				button.status:SetOrientation("HORIZONTAL")
				button.status:SetHeight(34)
				button.status:SetWidth(length)
				button.status:ClearAllPoints()
				button.status:SetPoint("TOPLEFT", button:GetName(), "TOPRIGHT")
				button.countdown:ClearAllPoints()
				button.countdown:SetPoint("LEFT", button.status:GetName(), "LEFT", 14, 0)
				button.spellName:Show()
			else
				x = (buttonIdx-1) * (width+margin) + outer_margin
				y = -outer_margin
				button.status:SetOrientation("VERTICAL")
				button.status:SetHeight(length)
				button.status:SetWidth(34)
				button.status:ClearAllPoints()
				button.status:SetPoint("BOTTOMLEFT", button:GetName(), "TOPLEFT")
				button.countdown:ClearAllPoints()
				button.countdown:SetPoint("BOTTOMLEFT", button.status:GetName(), "BOTTOMLEFT", 0, 14)
				button.spellName:Hide()
--				button.status:GetStatusBarTexture():SetTexCoord(1,0,0,0,1,1,0,1)
			end
			button:ClearAllPoints()
			button:SetPoint ("TOPLEFT", x, y)
			if locked then
				bar.texture:Hide()
			else
				bar.texture:Show()
			end
		else
			button:Hide()
		end
	end
	
	-- layout
	if width ~= 0 and height ~= 0 then
		local barWidth
		local barHeight
		if orientation == 1 then -- vertical
			barHeight = (height * buttonCount) + ((buttonCount - 1) * margin) + (outer_margin * 2)
			barWidth = width + (outer_margin * 2)
		else -- horizontal
			barHeight = height + (outer_margin * 2)
			barWidth = (width * buttonCount) + ((buttonCount - 1) * margin) + (outer_margin * 2)
		end
		ABT:DebugPrint ("barHeight="..barHeight.." barWidth="..barWidth)
		bar:SetHeight(barHeight)
		bar:SetWidth(barWidth)
	end
	ABT:DebugTrace ("SetBar")
end

function ABT:UpdatePlayerTargets()
ABT:DebugTrace ("UpdatePlayerTargets", true)
	ABT:InitButtonLoop()
	repeat 
		barIdx, buttonIdx = ABT:GetNextButton()
		ABT:DebugPrint ("UPT==bar"..NS(barIdx).."button"..NS(buttonIdx))
		if barIdx and buttonIdx then
			if (ABT:GetValue(barIdx, buttonIdx, "target") == ABT_PARTY) then
				local playerTarget = ABT:GetValue (barIdx, buttonIdx, "playerTarget")
				local unitId = ABT:FindPlayerUnitId (playerTarget)
				local button = ABT_Buttons[barIdx][buttonIdx]
				button:SetAttribute ("unit", unitId)
			end
		end
	until buttonIdx == nil
	ABT:GetDebuffs()
	ABT:DebugTrace ("UpdatePlayerTargets")
end

function ABT:FindPlayerUnitId (playerTarget)
	local unitId = nil
	ABT:DebugPrint ("FindPlayerUnitId looking for: "..NS(playerTarget))
	if playerTarget ~= nil then
		local raidNum = GetNumRaidMembers()
		if raidNum > 0 then
			for i = 1, raidNum do
				if UnitName("raid"..i) == playerTarget then
					unitId = "raid"..i
					break
				end
			end
		else
			local partyNum = GetNumPartyMembers()
			for i = 1, partyNum do
				if UnitName("party"..i) == playerTarget then
					unitId = "party"..i
					break
				end
			end
		end
	end
	ABT:DebugPrint ("FindPlayerUnitId found "..NS(playerTarget).." as "..NS(unitId))
	return unitId
end

function ABT:GetPartyList ()
	local _table = {}
	local playerName = nil
	local raidNum = GetNumRaidMembers()
	if raidNum > 0 then
		for i = 1, raidNum do
			playerName = UnitName("raid"..i)
			if playerName then _table[playerName] = playerName end
		end
	else
		for i = 1, 4 do
			playerName = UnitName("party"..i)
			if playerName then _table[playerName] = playerName end
		end
	end
	return _table
end

